/**
 * This program has been developed by students from the bachelor Computer Science at
 * Utrecht University within the Software Project course.
 * © Copyright Utrecht University (Department of Information and Computing Sciences)
 */

import { UseIsAuthorizedState } from '../../store/authSlice';
import Broker from '../broker';
import { ReceiveMessageI, SendMessageI, SendMessageWithSessionI } from '../types';
import BackendMessageReceiver from './BackendMessageReceiver';

/** The websockethandler creates a websocket and wait for messages send to the socket. */
export class WebSocketHandler implements BackendMessageReceiver {
  private static singletonInstance: WebSocketHandler;
  private webSocket: WebSocket | undefined;
  private url: string;
  private connected: boolean;
  private authHeader: UseIsAuthorizedState | undefined;
  private saveStateID: string | undefined;

  /** @param domain The domain to make the websocket connection with. */
  public constructor(domain: string) {
    this.url = domain;
    this.connected = false;
  }

  public static instance(): WebSocketHandler {
    if (!this.singletonInstance) this.singletonInstance = new WebSocketHandler(import.meta.env.BACKEND_WSS_URL);
    return this.singletonInstance;
  }

  public useAuth(authHeader: UseIsAuthorizedState): WebSocketHandler {
    this.authHeader = authHeader;
    return this;
  }

  public useSaveStateID(saveStateID: string): WebSocketHandler {
    this.saveStateID = saveStateID;
    return this;
  }

  /**
   * Create a websocket to the given URL.
   * @param {string} URL is the URL to which the websocket connection is opened.
   */
  public connect(onOpen: () => void): void {
    // If there already is already a current websocket connection, close it first.
    if (this.webSocket) this.close();

    const params = new URLSearchParams(window.location.search);
    // Most of these parameters are only really used in DEV
    if (this.authHeader?.userID) params.set('userID', this.authHeader?.userID ?? '');
    if (this.authHeader?.roomID) params.set('roomID', this.authHeader?.roomID ?? '');
    if (this.saveStateID) params.set('saveStateID', this.saveStateID ?? '');
    if (this.authHeader?.sessionID) params.set('sessionID', this.authHeader?.sessionID ?? '');
    if (this.authHeader?.jwt) params.set('jwt', this.authHeader?.jwt ?? '');
    this.webSocket = new WebSocket(this.url + '?' + params.toString());
    this.webSocket.onopen = () => {
      this.connected = true;
      onOpen();
    };
    this.webSocket.onmessage = this.onWebSocketMessage;
    this.webSocket.onerror = this.onError;
    this.webSocket.onclose = this.onClose;
  }

  public sendMessage(message: SendMessageI): void {
    console.debug('%cSending WS message: ', 'background: #222; color: #bada55', message);
    let fullMessage = message as SendMessageWithSessionI;
    fullMessage.sessionID = this.authHeader?.sessionID ?? '';
    if (message.body && typeof message.body !== 'string') {
      fullMessage.body = JSON.stringify(message.body);
    }

    if (this.webSocket && this.connected && this.webSocket.readyState === 1) this.webSocket.send(JSON.stringify(fullMessage));
    else
      this.connect(() => {
        if (this.webSocket && this.connected && this.webSocket.readyState === 1) this.webSocket.send(JSON.stringify(fullMessage));
      });
  }

  /** Closes the current websocket connection. */
  public close = (): void => {
    if (this.webSocket) this.webSocket.close();
    this.connected = false;
    this.webSocket = undefined;
  };

  /** @returns A boolean which indicates if there currently is a socket connection. */
  public isConnected = (): boolean => {
    return this.connected;
  };

  public attemptReconnect = () => {
    console.warn('Attempting to reconnect WS');

    if (!this.connected || !this.webSocket || this.webSocket.readyState !== 1) {
      this.connect(() => {
        setTimeout(() => WebSocketHandler.instance().attemptReconnect(), 5000);
      });
    } else {
      console.log('WS reconnected', this.webSocket?.readyState, this.connected);
    }
  };

  /**
   * Websocket connection close event handler.
   * @param {any} event Contains the event data.
   */
  private onClose(event: any): void {
    console.warn('WS connection was closed from the server side', event.data);
    if (this.webSocket) this.webSocket.close();
    this.connected = false;
    this.webSocket = undefined;
    setTimeout(() => WebSocketHandler.instance().attemptReconnect(), 5000);
  }

  /**
   * Websocket connection message event handler. Called if a new message is received through the socket.
   * @param {any} event Contains the event data.
   */
  public onWebSocketMessage = (event: MessageEvent<any>) => {
    let data: ReceiveMessageI = JSON.parse(event.data);
    Broker.instance().publish(data.value, data.type);
  };

  /**
   * Websocket connection error event handler.
   * @param {any} event contains the event data.
   */
  private onError(event: any): void {
    console.error('WS error', event);
  }
}
