import { createSlice, PayloadAction } from '@reduxjs/toolkit';
import type { RootState } from './store';
import { DatabaseStatus, SaveStateI } from '../api/wsState';
import { getParam, URLParams } from '../api/url';

/** Message format of the error message from the backend */
export type ErrorMessage = {
  errorStatus: any;
  queryID: any;
};

/** Cache type */
export type SessionCacheI = {
  currentSaveState?: string; // id of the current save state
  saveStates: Record<string, SaveStateI>;
  testedSaveState: Record<string, DatabaseStatus>;
};

// Define the initial state using that type
export const initialState: SessionCacheI = {
  currentSaveState: undefined,
  saveStates: {},
  testedSaveState: {},
};

export const sessionSlice = createSlice({
  name: 'session',
  initialState,
  reducers: {
    selectSaveState(state, action: PayloadAction<string | undefined>) {
      if (action.payload !== undefined && action.payload in state.saveStates) {
        state.currentSaveState = action.payload;
      } else {
        state.currentSaveState = undefined;
      }
    },
    updateSelectedSaveState(state, action: PayloadAction<SaveStateI>) {
      if (state.currentSaveState === action.payload.id)
        //@ts-ignore
        state.saveStates = {
          ...state.saveStates,
          [state.currentSaveState]: action.payload,
        };
    },
    updateSaveStateList(state, action: PayloadAction<SaveStateI[]>) {
      // Does NOT clear the states, just adds in new data
      let newState: Record<string, SaveStateI> = {};
      action.payload.forEach((ss) => {
        newState[ss.id] = ss;
      });
      state.saveStates = { ...state.saveStates, ...newState };

      const paramSaveState = getParam(URLParams.saveState);
      if (!state.currentSaveState) {
        if (paramSaveState && paramSaveState in state.saveStates) {
          state.currentSaveState = paramSaveState;
        } else if (Object.keys(state.saveStates).length > 0) {
          state.currentSaveState = Object.keys(state.saveStates)[0];
        } else state.currentSaveState = undefined;
      }
    },
    setSaveStateList(state, action: PayloadAction<SaveStateI[]>) {
      // Clears the states and puts in new data
      let newState: Record<string, SaveStateI> = {};
      action.payload.forEach((ss) => {
        newState[ss.id] = ss;
      });
      state.saveStates = newState;

      const paramSaveState = getParam(URLParams.saveState);
      if (!state.currentSaveState) {
        if (paramSaveState && paramSaveState in state.saveStates) {
          state.currentSaveState = paramSaveState;
        } else if (Object.keys(state.saveStates).length > 0) {
          state.currentSaveState = Object.keys(state.saveStates)[0];
        } else state.currentSaveState = undefined;
      }
    },
    addSaveState(state, action: PayloadAction<SaveStateI>) {
      if (state.saveStates === undefined) state.saveStates = {};
      state.saveStates[action.payload.id] = action.payload;
      state.currentSaveState = action.payload.id;
    },
    deleteSaveState(state, action: PayloadAction<string>) {
      delete state.saveStates[action.payload];
      if (state.currentSaveState === action.payload) {
        if (Object.keys(state.saveStates).length > 0) state.currentSaveState = Object.keys(state.saveStates)[0];
        else state.currentSaveState = undefined;
      }
    },
    testedSaveState(state, action: PayloadAction<string>) {
      state.testedSaveState = { ...state.testedSaveState, [action.payload]: DatabaseStatus.tested };
    },
  },
});

export const {
  selectSaveState,
  deleteSaveState,
  updateSaveStateList,
  setSaveStateList,
  addSaveState,
  testedSaveState,
  updateSelectedSaveState,
} = sessionSlice.actions;

// Other code such as selectors can use the imported `RootState` type
export const sessionCacheState = (state: RootState) => state.sessionCache;

export default sessionSlice.reducer;
