import { useAppDispatch, useML, useSearchResultData } from '@graphpolaris/shared/lib/data-access';
import { dataColors, tailwindColors } from 'config';
import { Viewport } from 'pixi-viewport';
import { Application, Color, Container, FederatedPointerEvent, Graphics, IPointData, Point, Text } from 'pixi.js';
import { useEffect, useImperativeHandle, useRef, useState } from 'react';
import { GraphType, LinkType, NodeType } from '../Types';
import { NLPopup } from './NLPopup';
import { hslStringToHex, nodeColor } from './utils';

import { CytoscapeLayout, GraphologyLayout, LayoutFactory, Layouts } from '@graphpolaris/shared/lib/graph-layout';
import { Cull } from '@pixi-essentials/cull';
import { MultiGraph } from 'graphology';
import { NodeLinkProps } from '../nodelinkvis';

type Props = {
  onClick: (node: NodeType, pos: IPointData) => void;
  // onHover: (data: { node: NodeType; pos: IPointData }) => void;
  // onUnHover: (data: { node: NodeType; pos: IPointData }) => void;
  selectedNodes: NodeType[];
  setSelectedNodes: (nodes: NodeType[]) => void;
  currentShortestPathEdges?: LinkType[];
  highlightedLinks?: LinkType[];
  graph?: GraphType;
  localConfig: NodeLinkProps;
};

const app = new Application({
  background: 0xffffff,
  antialias: true,
  autoDensity: true,
  eventMode: 'auto',
  resolution: window.devicePixelRatio || 1,
});
const nodeLayer = new Container();
const nodeLabelLayer = new Container();
const linkLayer = new Container();
const linkLabelLayer = new Container();
const lassoSelectionLayer = new Container();

//////////////////
// MAIN COMPONENT
//////////////////

export const NLPixi = (props: Props) => {
  const [quickPopup, setQuickPopup] = useState<{ node: NodeType; pos: IPointData } | undefined>();
  const [popups, setPopups] = useState<{ node: NodeType; pos: IPointData }[]>([]);

  const nodeMap = useRef(new Map<string, Graphics>());
  const nodeLabelMap = useRef(new Map<string, Text>());
  const linkMap = useRef(new Map<string, Graphics>());
  const linkLabelMap = useRef(new Map<string, Text>());
  const viewport = useRef<Viewport>();
  const ref = useRef<HTMLDivElement>(null);
  const isSetup = useRef(false);
  const ml = useML();
  const dragging = useRef<{ node: NodeType; gfx: Graphics | Text } | null>(null);
  const onlyClicked = useRef(false);
  const dispatch = useAppDispatch();
  const searchResults = useSearchResultData();

  let layoutAlgorithm: CytoscapeLayout | GraphologyLayout = new LayoutFactory().createLayout(Layouts.DAGRE);

  const cull = new Cull();
  let cullDirty = useRef(true);

  const [config, setConfig] = useState({
    width: 1000,
    height: 1000,

    LAYOUT_ALGORITHM: Layouts.FORCEATLAS2WEBWORKER,

    NODE_RADIUS: 5,
    NODE_BORDER_LINE_WIDTH: 1.0,
    NODE_BORDER_LINE_WIDTH_SELECTED: 1.0, // if selected and normal width are different the thicker line will be still in the gfx
    NODE_BORDER_COLOR_DEFAULT: dataColors.black,
    NODE_BORDER_COLOR_SELECTED: dataColors.orange[60],

    LINE_COLOR_DEFAULT: dataColors.neutral[20],
    LINE_COLOR_SELECTED: tailwindColors.entity[400],
    LINE_COLOR_ML: dataColors.blue[60],
    LINE_WIDTH_DEFAULT: 0.8,
  });

  useEffect(() => {
    if (props.graph) {
      setup();
    }
  }, [config]);

  useEffect(() => {
    setConfig((lastConfig) => {
      return {
        ...lastConfig,
        LAYOUT_ALGORITHM: (props.localConfig.layoutAlgorithm as Layouts) || lastConfig.LAYOUT_ALGORITHM,
      };
    });
  }, [props.localConfig]);

  const imperative = useRef<any>(null);

  useImperativeHandle(imperative, () => ({
    onDragStart(node: NodeType, gfx: Graphics | Text) {
      if (viewport.current) viewport.current.pause = true;
      dragging.current = { node, gfx };
      onlyClicked.current = true;
    },

    onDragMove(movementX: number, movementY: number) {
      if (dragging.current) {
        onlyClicked.current = false;
        // if (popups.length > 0) setPopups([]);
        if (quickPopup) setQuickPopup(undefined);
        const idx = popups.findIndex((p) => p.node.id === dragging.current?.node.id);
        if (idx >= 0) {
          const p = popups[idx];
          p.pos.x += movementX / (viewport.current?.scaled || 1);
          p.pos.y += movementY / (viewport.current?.scaled || 1);
          popups[idx] = p;
          setPopups([...popups]);
        }

        if (!dragging.current.node.fx) dragging.current.node.fx = dragging.current.node.x || 0;
        if (!dragging.current.node.fy) dragging.current.node.fy = dragging.current.node.y || 0;
        dragging.current.node.fx += movementX / (viewport.current?.scaled || 1);
        dragging.current.node.fy += movementY / (viewport.current?.scaled || 1);
        // force.simulation.alpha(0.1).restart();
      }
    },

    onDragEnd() {
      if (dragging.current) {
        // dragging.current.node.fx = null;
        // dragging.current.node.fy = null;
        if (viewport.current) viewport.current.pause = false;
        if (onlyClicked.current) {
          onlyClicked.current = false;

          if (popups.filter((d) => d.node.id === dragging.current?.node.id).length > 0) {
            setPopups(popups.filter((p) => p.node.id !== dragging.current?.node.id));
          } else {
            console.log('clicked', popups);
            setPopups([...popups, { node: dragging.current.node, pos: toGlobal(dragging.current.node) }]);
          }

          props.onClick(dragging.current.node, toGlobal(dragging.current.node));
        }
        this.onHover(dragging.current.node);
        dragging.current = null;
      }
    },
    onHover(node: NodeType) {
      if (viewport?.current && !viewport?.current?.pause && node && popups.filter((p) => p.node.id === node.id).length === 0) {
        setQuickPopup({ node, pos: toGlobal(node) });
      }
    },
    onUnHover() {
      setQuickPopup(undefined);
    },
    onPan() {
      setPopups([]);
    },
  }));

  // useEffect(() => {
  //   app.renderer.resize(props.windowSize.width, props.windowSize.height);
  //   app.render();
  // }, [props.windowSize]);

  function resize() {
    const width = ref?.current?.clientWidth || 1000;
    const height = ref?.current?.clientHeight || 1000;
    app.renderer.resize(width, height);
    if (viewport.current) {
      viewport.current.screenWidth = width;
      viewport.current.worldWidth = width;
      viewport.current.worldHeight = height;
      viewport.current.screenHeight = height;
    }
    app.render();
  }

  useEffect(() => {
    if (!ref.current) return;
    const resizeObserver = new ResizeObserver(() => {
      resize();
    });
    resizeObserver.observe(ref.current);
    return () => resizeObserver.disconnect(); // clean up
  }, []);

  useEffect(() => {
    if (ref.current && ref.current.children.length === 0) {
      ref.current.appendChild(app.view as HTMLCanvasElement);
      resize();
    }
  }, [ref]);

  function toGlobal(node: NodeType): IPointData {
    if (viewport?.current) {
      // const rect = ref.current?.getBoundingClientRect();
      const rect = { x: 0, y: 0 };
      const x = (rect?.x || 0) + (node.x || 0);
      const y = (rect?.y || 0) + (node.y || 0);
      return viewport.current.toScreen(x, y);
    } else return { x: 0, y: 0 };
  }

  function onDragStart(event: FederatedPointerEvent, node: NodeType, gfx: Graphics | Text) {
    event.stopPropagation();
    updateVisibility();
    if (imperative.current) imperative.current.onDragStart(node, gfx);
  }

  function onDragMove(event: FederatedPointerEvent) {
    event.stopPropagation();
    updateVisibility();
    if (imperative.current) imperative.current.onDragMove(event.movementX, event.movementY);
  }

  function onDragEnd(event: FederatedPointerEvent) {
    event.stopPropagation();
    updateVisibility();
    if (imperative.current) imperative.current.onDragEnd();
  }

  const updateNode = (node: NodeType) => {
    const gfx = nodeMap.current.get(node.id);
    if (!gfx) return;

    const lineColor = node.isShortestPathSource
      ? tailwindColors.entity[950]
      : node.isShortestPathTarget
      ? tailwindColors.relation[950]
      : node.selected
      ? config.NODE_BORDER_COLOR_SELECTED
      : config.NODE_BORDER_COLOR_DEFAULT;
    const lineWidth = node.selected ? config.NODE_BORDER_LINE_WIDTH_SELECTED : config.NODE_BORDER_LINE_WIDTH;
    gfx.lineStyle(lineWidth, new Color(hslStringToHex(lineColor)));

    if (node?.cluster) {
      gfx.beginFill(node.cluster >= 0 ? nodeColor(node.cluster) : 0x000000);
    } else gfx.beginFill(nodeColor(node.type));

    gfx.drawCircle(0, 0, Math.max(node.radius || 5, 5));
    gfx.endFill();

    gfx.position.set(node.x, node.y);

    gfx.off('mouseover');
    gfx.off('mousedown');
    gfx.on('mouseover', (e) => {
      e.stopPropagation();
      if (imperative.current) imperative.current.onHover(node);
    });
    gfx.on('mouseout', (e) => {
      e.stopPropagation();
      if (imperative.current) imperative.current.onUnHover();
    });
    gfx.on('mousedown', (e) => onDragStart(e, node, gfx));
  };

  const updateNodeLabel = (node: NodeType) => {
    const label = nodeLabelMap.current.get(node.id);
    if (!label) return;

    if (node?.cluster) {
      label.tint = node.cluster >= 0 ? nodeColor(node.cluster) : 0x000000;
    } else label.tint = nodeColor(node.type);

    label.position.set(node.x, node.y);

    label.off('mouseover');
    label.off('mousedown');
    label.on('mouseover', (e) => {
      e.stopPropagation();
      if (imperative.current) imperative.current.onHover(node);
    });
    label.on('mouseout', (e) => {
      e.stopPropagation();
      if (imperative.current) imperative.current.onUnHover();
    });
    label.on('mousedown', (e) => onDragStart(e, node, label));

    const rect = nodeLabelLayer.getChildByName('nodelabelrect_' + node.id);
    if (rect && node.x && node.y) {
      console.log('rect', rect, node.x, node.y);
      rect.position.set(node.x - 50, node.y - 10);
    }

    // if (!item.position) {
    //   item.position = new Point(node.x, node.y);
    // } else {
    //   item.position.set(node.x, node.y);
    // }
    // Update attributes position if they exist
    // if (node.gfxAttributes) {
    //   const x = node.x - node.gfxAttributes.width / 2;
    //   const y = node.y - node.gfxAttributes.height - 20;
    //   if (!node.gfxAttributes?.position) node.gfxAttributes.position = new Point(x, y);
    //   else {
    //     node.gfxAttributes.position.set(x, y);
    //   }
    // }
  };

  const createNode = (node: NodeType, selected?: boolean) => {
    // check if node is already drawn, and if so, delete it
    if (node && node?.id && nodeMap.current?.has(node.id)) {
      nodeMap.current.delete(node.id);
    }
    // Do not draw node if it has no position
    if (node.x === undefined || node.y === undefined) return;
    const gfx = new Graphics();
    nodeMap.current.set(node.id, gfx);
    nodeLayer.addChild(gfx);
    node.selected = selected;

    updateNode(node);
    gfx.name = 'node_' + node.id;
    gfx.eventMode = 'dynamic';

    return gfx;
  };

  const createNodeLabel = (node: NodeType, selected?: boolean) => {
    // check if node is already drawn, and if so, delete it
    if (node && node?.id && nodeLabelMap.current?.has(node.id)) {
      nodeLabelMap.current.delete(node.id);
    }
    // Do not draw node if it has no position
    if (node.x === undefined || node.y === undefined) return;

    // Create the semi-transparent white rectangle
    // const rect = new Graphics();
    // rect.beginFill(0xffffff, 0.5); // Set fill color to white with 50% opacity
    // rect.drawRect(-50, -10, 100, 20); // Set rectangle dimensions and position
    // rect.endFill();
    // rect.name = 'nodelabelrect_' + node.id;

    // Create the label
    const label = new Text(node.displayInfo || node.id, {
      fontFamily: 'Roboto Light',
      fontSize: 200,
      fill: '#000000',
    });
    label.scale.set(0.03);
    label.anchor.set(0.5);
    label.position.set(node.x, node.y + 20);

    // Add the rectangle and label to the nodeLabelLayer
    // nodeLabelLayer.addChild(rect);
    nodeLabelLayer.addChild(label);

    nodeLabelMap.current.set(node.id, label);
    node.selected = selected;

    updateNodeLabel(node);
    label.name = 'nodelabel_' + node.id;
    label.eventMode = 'dynamic';

    return label;
  };

  // /** UpdateRadius works just like UpdateColors, but also applies radius*/
  // const UpdateRadius = (graph: GraphType, radius: number) => {
  //   // update for each node in graph
  //   graph.nodes.forEach((node: NodeType) => {
  //     createNode(node);
  //   });
  // };

  const createLink = (link: LinkType) => {
    if (linkMap.current.has(link.id)) {
      linkMap.current.delete(link.id);
    }
    const gfx = new Graphics();
    gfx.name = 'link_' + link.id;
    linkMap.current.set(link.id, gfx);
    updateLink(link, gfx);
    linkLayer.addChild(gfx);
    return gfx;
  };

  const updateLink = (link: LinkType, gfx: Graphics) => {
    if (!props.graph) return;

    const _source = link.source;
    const _target = link.target;

    if (!_source || !_target) {
      console.log('source or target not found', _source, _target);
      return;
    }

    let sourceId = '';
    let targetId = '';
    let source: NodeType | undefined;
    let target: NodeType | undefined;

    if (typeof _source === 'string') {
      sourceId = link.source as string;
      targetId = link.target as string;
      source = nodeMap.current.get(sourceId) as NodeType | undefined;
      target = nodeMap.current.get(targetId) as NodeType | undefined;
    } else {
      source = link.source as NodeType;
      target = link.target as NodeType;
      sourceId = source.id;
      targetId = target.id;
    }
    if (!source || !target) {
      console.error('source or target not found', source, target, sourceId, targetId);
      return;
    }

    if (gfx) {
      // let color = link.color || 0x000000;
      let color = config.LINE_COLOR_DEFAULT;
      let style = config.LINE_WIDTH_DEFAULT;
      let alpha = link.alpha || 1;
      if (link.mlEdge) {
        color = config.LINE_COLOR_ML;
        if (link.value > ml.communityDetection.jaccard_threshold) {
          style = link.value * 1.8;
        } else {
          style = 0;
          alpha = 0.2;
        }
      } else if (props.highlightedLinks && props.highlightedLinks.includes(link)) {
        if (link.mlEdge && ml.communityDetection.jaccard_threshold) {
          if (link.value > ml.communityDetection.jaccard_threshold) {
            color = dataColors.magenta[50];
            // 0xaa00ff;
            style = link.value * 1.8;
          }
        } else {
          color = dataColors.red[70];
          // color = 0xff0000;
          style = 1.0;
        }
      } else if (props.currentShortestPathEdges && props.currentShortestPathEdges.includes(link)) {
        color = dataColors.green[50];
        // color = 0x00ff00;
        style = 3.0;
      }

      gfx.clear();
      gfx.beginFill();
      gfx
        .lineStyle(style, hslStringToHex(color), alpha)
        .moveTo(source.x || 0, source.y || 0)
        .lineTo(target.x || 0, target.y || 0);
      gfx.endFill();
    } else {
      throw Error('Link not found');
    }
  };

  const createLinkLabel = (link: LinkType) => {
    if (linkLabelMap.current.has(link.id)) {
      linkLabelMap.current.delete(link.id);
    }
    // Create the label
    const label = new Text(link.id, {
      fontFamily: 'Roboto Light',
      fontSize: 200,
      fill: '#000000',
    });
    label.name = 'linklabel_' + link.id;
    label.scale.set(0.03);
    label.anchor.set(0.5);
    // label.position.set(link.x, link.y + 20);
    linkLabelMap.current.set(link.id, label);
    linkLabelLayer.addChild(label);

    updateLinkLabel(link);
    label.name = 'linklabel_' + link.id;
    label.eventMode = 'dynamic';

    return label;
  };

  const updateLinkLabel = (link: LinkType) => {
    const label = linkLabelMap.current.get(link.id);
    if (!label || !props.graph) return;

    const _source = link.source;
    const _target = link.target;

    if (!_source || !_target) {
      console.log('source or target not found', _source, _target);
      return;
    }

    let sourceId = '';
    let targetId = '';
    let source: NodeType | undefined;
    let target: NodeType | undefined;

    if (typeof _source === 'string') {
      sourceId = link.source as string;
      targetId = link.target as string;
      source = nodeMap.current.get(sourceId) as NodeType | undefined;
      target = nodeMap.current.get(targetId) as NodeType | undefined;
    } else {
      source = link.source as NodeType;
      target = link.target as NodeType;
      sourceId = source.id;
      targetId = target.id;
    }
    if (!source || !target) {
      console.error('source or target not found', source, target, sourceId, targetId);
      return;
    }

    if (!label) throw Error('LinkLabel not found');

    // label.tint = hexToNumber(tailwindColors.relation[900]);

    // Define start and end points
    let startPoint = { x: source.x || 0, y: source.y || 0 };
    let endPoint = { x: target.x || 0, y: target.y || 0 };

    // Calculate midpoint
    let midPoint = {
      x: (startPoint.x + endPoint.x) / 2,
      y: (startPoint.y + endPoint.y) / 2,
    };

    // Calculate angle in radians
    let angle = Math.atan2(endPoint.y - startPoint.y, endPoint.x - startPoint.x);

    // If the angle is less than 0, add π to it to ensure the text is always right side up
    if (endPoint.x < startPoint.x) {
      angle += Math.PI;
    }

    // Rotate the text
    label.rotation = angle;

    // Position the text to be in the middle of the line
    label.position.set(midPoint.x, midPoint.y);

    // // Create the background rectangle
    // const rect = new Graphics();
    // rect.beginFill(0x000000, 0.5); // Set the fill color and transparency
    // rect.drawRect(-label.width / 2, -label.height / 2, label.width, label.height);
    // rect.endFill();
    // label.addChild(rect);

    return label;
  };

  useEffect(() => {
    if (props.graph && ref.current && ref.current.children.length > 0 && imperative.current) {
      if (!isSetup.current) setup();
      else update();
      // setup();
    }
  }, [props.graph]);

  useEffect(() => {
    if (!props.graph || searchResults === undefined) return;

    if (searchResults.nodes && props.graph.nodes) {
      props.graph.nodes.forEach((node: NodeType) => {
        const gfx = nodeMap.current.get(node.id);
        if (!gfx) return;
        const isNodeInSearchResults = searchResults.nodes.some((resultNode) => resultNode.id === node.id);
        gfx.alpha = isNodeInSearchResults || searchResults.nodes.length === 0 ? 1 : 0.05;
      });
    }

    if (searchResults.edges && props.graph.links) {
      props.graph.links.forEach((link: LinkType) => {
        const gfx = linkMap.current.get(link.id);
        if (!gfx) return;
        const isLinkInSearchResults = searchResults.edges.some((resultEdge) => resultEdge.id === link.id);
        gfx.alpha = isLinkInSearchResults || searchResults.edges.length === 0 ? 1 : 0.05;
      });
    }
  }, [searchResults]);

  const tick = (delta: number) => {
    if (props.graph) {
      // updateVisibility();

      if (!layoutAlgorithm) return;
      // const layoutGraphModel = layoutAlgorithm.getFullLayoutUpdate();
      // if (!layoutGraphModel) return;
      // layoutGraphModel.nodes().forEach((nodeID: string) => {
      //   const position = layoutGraphModel.getNodeAttributes(nodeID);

      //   if (!position || position.x === undefined || position.y === undefined) return new Error('position not found for node ID in layout');

      //   const gfx = nodeMap.current.get(nodeID);
      //   if (!gfx) return;
      //   gfx.position.copyFrom(position as IPointData);

      //   const label = nodeLabelMap.current.get(nodeID);
      //   if (!label) return;
      //   label.position.copyFrom(position as IPointData);
      //   label.position.y += gfx.height;

      //   console.log('tick');
      // });

      props.graph.nodes.forEach((node: NodeType) => {
        if (!layoutAlgorithm) return;
        const gfx = nodeMap.current.get(node.id);

        const position = layoutAlgorithm.getNodePosition(node.id);
        if (node.x === position.x && node.y === position.y) return;

        if (layoutAlgorithm.provider === 'Graphology') {
          // this is a dirty hack to fix the graphology layout being out of bounds
          node.x = position.x + app.renderer.width / 2;
          node.y = position.y + app.renderer.height / 2;
        } else {
          node.x = position.x;
          node.y = position.y;
        }

        if (!gfx || node.x === undefined || node.y === undefined) return;
        gfx.position.copyFrom(node as IPointData);

        const label = nodeLabelMap.current.get(node.id);
        if (!label || node.x === undefined || node.y === undefined) return;
        label.position.copyFrom(node as IPointData);
        label.position.y += gfx.height;
      });

      // Update forces of the links
      props.graph.links.forEach((link: any) => {
        if (linkMap.current && !!linkMap.current.has(link.id)) {
          updateLink(link, linkMap.current.get(link.id) as Graphics);
          updateLinkLabel(link);
        }
      });
    }
  };

  const updateVisibility = () => {
    if (!viewport.current) return;
    // culling
    if (viewport.current.dirty || cullDirty.current) {
      cull.cull(app.renderer.screen);
      // console.log(
      //   'culling',
      //   [...nodeLabelLayer.children].filter((x) => x.visible === false).length,
      //   [...nodeLayer.children].filter((x) => x.visible === false).length,
      //   [...linkLabelLayer.children].filter((x) => x.visible === false).length,
      //   [...linkLayer.children].filter((x) => x.visible === false).length
      // );

      viewport.current.dirty = false;
      cullDirty.current = false;
    }

    // levels of detail
    const zoom = viewport.current.scale.x || 1;
    const zoomSteps = [0.7, 1.5, 2, Infinity];
    const zoomStep = zoomSteps.findIndex((zoomStep) => zoom <= zoomStep);

    // console.log('zoomStep', zoomStep, zoom);

    nodeLabelLayer.children.forEach((label) => {
      label.visible = zoomStep >= 2;
    });

    linkLabelLayer.children.forEach((label) => {
      label.visible = zoomStep >= 2;
    });

    linkLayer.children.forEach((link) => {
      link.visible = zoomStep >= 1;
    });
  };

  const update = (forceClear = false) => {
    setPopups([]);
    if (!props.graph || !ref.current) return;

    if (props.graph) {
      if (forceClear) {
        nodeMap.current.clear();
        nodeLabelMap.current.clear();
        linkMap.current.clear();
        linkLabelMap.current.clear();
        nodeLayer.removeChildren();
        nodeLabelLayer.removeChildren();
        linkLayer.removeChildren();
        linkLabelLayer.removeChildren();
      }

      nodeMap.current.forEach((gfx, id) => {
        if (!props.graph?.nodes?.find((node) => node.id === id)) {
          nodeLayer.removeChild(gfx);
          gfx.destroy();
          nodeMap.current.delete(id);
        }
      });

      nodeLabelMap.current.forEach((gfx, id) => {
        if (!props.graph?.nodes?.find((node) => node.id === id)) {
          nodeLabelLayer.removeChild(gfx);
          gfx.destroy();
          nodeLabelMap.current.delete(id);
        }
      });

      linkMap.current.forEach((gfx, id) => {
        if (!props.graph?.links?.find((link) => link.id === id)) {
          linkLayer.removeChild(gfx);
          gfx.destroy();
          linkMap.current.delete(id);
        }
      });

      linkLabelMap.current.forEach((gfx, id) => {
        if (!props.graph?.links?.find((link) => link.id === id)) {
          linkLabelLayer.removeChild(gfx);
          gfx.destroy();
          linkLabelMap.current.delete(id);
        }
      });

      props.graph.nodes.forEach((node: NodeType) => {
        if (!forceClear && nodeMap.current.has(node.id)) {
          const old = nodeMap.current.get(node.id);
          node.x = old?.x || node.x;
          node.y = old?.y || node.y;
          updateNode(node);
        } else {
          createNode(node);
        }
      });

      props.graph.nodes.forEach((node: NodeType) => {
        if (!forceClear && nodeLabelMap.current.has(node.id)) {
          const old = nodeLabelMap.current.get(node.id);
          node.x = old?.x || node.x;
          node.y = old?.y || node.y;
          updateNodeLabel(node);
        } else {
          createNodeLabel(node);
        }
      });

      props.graph.links.forEach((link: LinkType) => {
        if (!forceClear && linkMap.current.has(link.id)) {
          const gfx = linkMap.current.get(link.id);
          if (gfx) updateLink(link, gfx);
        } else createLink(link);
      });

      props.graph.links.forEach((link: LinkType) => {
        if (!forceClear && linkLabelMap.current.has(link.id)) {
          const linkLabel = linkLabelMap.current.get(link.id);
          if (linkLabel) updateLinkLabel(link);
        } else {
          createLinkLabel(link);
        }
      });

      cull.clear();
      cull.addAll(nodeLabelLayer.children);
      cull.addAll(nodeLayer.children);
      cull.addAll(linkLabelLayer.children);
      cull.addAll(linkLayer.children);
      cullDirty.current = true;

      // force.startSimulation(props.graph, ref.current.getBoundingClientRect());
      // force.simulation.on('tick', () => {});

      updateVisibility();

      app.ticker.add(tick);
    }
  };

  /**
   * SetNodeGraphics is an initializing function. It attaches the nodes and links to the simulation.
   * It creates graphic objects and adds these to the PIXI containers. It also clears both of these of previous nodes and links.
   * @param graph The graph returned from the database and that is parsed into a nodelist and edgelist.
   */
  const setup = () => {
    lassoSelectionLayer.removeChildren();

    nodeLayer.removeChildren();
    nodeLabelLayer.removeChildren();
    linkLayer.removeChildren();
    linkLabelLayer.removeChildren();

    app.stage.removeChildren();

    if (!props.graph) throw Error('Graph is undefined');

    const size = ref.current?.getBoundingClientRect();
    viewport.current = new Viewport({
      screenWidth: size?.width || 1000,
      screenHeight: size?.height || 1000,
      worldWidth: size?.width || 1000,
      worldHeight: size?.height || 1000,
      stopPropagation: true,
      events: app.renderer.events, // the interaction module is important for wheel to work properly when renderer.view is placed or scaled
    });

    app.stage.addChild(viewport.current);
    // activate plugins
    // viewport.current.drag().pinch().wheel().animate({ time: 1500, scale: 2.5, ease: 'easeInOutSine' }).decelerate({ friction: 0.75 });
    viewport.current.drag().pinch().wheel().animate({}).decelerate({ friction: 0.75 });

    // add topmost layer last to the list
    viewport.current.addChild(linkLayer);
    viewport.current.addChild(linkLabelLayer);
    viewport.current.addChild(nodeLayer);
    viewport.current.addChild(nodeLabelLayer);
    viewport.current.addChild(lassoSelectionLayer);

    viewport.current.on('drag-start', (event) => {
      imperative.current.onPan();
      updateVisibility();
    });
    viewport.current.on('wheel', (event) => {
      updateVisibility();
    });

    // app.stage.eventMode = 'dynamic';
    // app.stage.on('mouseup', onDragEnd);
    // app.stage.on('pointerup', onDragEnd);
    // app.stage.on('mousemove', onDragMove);
    // app.stage.on('mouseup', onDragEnd);

    app.ticker.add(tick);

    nodeMap.current.clear();
    nodeLabelMap.current.clear();
    linkMap.current.clear();
    linkLabelMap.current.clear();

    setupLassoSelection();
    setupLayout();

    // console.log('after layouting', graphologyGraph,);
    // graphologyGraph.forEachNode((node, attr) => {
    //   console.log('node', node, attr);
    // });

    update();
    isSetup.current = true;
  };

  const setupLassoSelection = () => {
    const lassoSelection = new Graphics();
    lassoSelectionLayer.addChild(lassoSelection);

    const MAX_LASSO_POINTS = 800;

    let isDragging = false;
    let startX = 0;
    let startY = 0;
    let points: Array<{ x: number; y: number }> = [];

    const startLassoSelection = (event: FederatedPointerEvent) => {
      resetAllSelections();

      if (!event.shiftKey) return;

      if (!viewport.current) throw Error('Viewport is undefined');

      viewport.current.plugins.pause('drag');
      if (!event.shiftKey) return;
      isDragging = true;
      const mousePoint = viewport.current.toWorld(event.global.x, event.global.y);
      startX = mousePoint.x;
      startY = mousePoint.y;
      points = [{ x: mousePoint.x, y: mousePoint.y }]; // Reset points array and add the first point
    };

    const updateLassoSelection = (event: FederatedPointerEvent) => {
      if (!event.shiftKey) return;
      if (!isDragging) return;

      if (!viewport.current) throw Error('Viewport is undefined');

      const mousePoint = viewport.current.toWorld(event.global.x, event.global.y);
      const currentX = mousePoint.x;
      const currentY = mousePoint.y;

      if (points.length < MAX_LASSO_POINTS) {
        points.push({ x: currentX, y: currentY }); // Add current point to the points array
      }

      lassoSelection.clear();
      lassoSelection.lineStyle(1, 0x000000);
      lassoSelection.beginFill(0x000000, 0.1);
      lassoSelection.drawPolygon(points.map((point) => new Point(point.x, point.y))); // Draw a polygon using the points
      lassoSelection.endFill();

      const checkNodesWithinPolygon = () => {
        if (!props.graph) return;

        function isPointInPolygon(point: { x: number; y: number }, polygon: Array<{ x: number; y: number }>) {
          let inside = false;
          for (let i = 0, j = polygon.length - 1; i < polygon.length; j = i++) {
            const xi = polygon[i].x,
              yi = polygon[i].y;
            const xj = polygon[j].x,
              yj = polygon[j].y;

            const intersect = yi > point.y !== yj > point.y && point.x < ((xj - xi) * (point.y - yi)) / (yj - yi) + xi;
            if (intersect) inside = !inside;
          }
          return inside;
        }

        props.graph.nodes.forEach((node) => {
          // const nodePosition = toGlobal(node);
          if (node.x === undefined || node.y === undefined) return;
          const nodePosition = { x: node.x, y: node.y };
          const isWithinPolygon = isPointInPolygon(nodePosition, points);

          if (isWithinPolygon) {
            node.selected = true;
          } else {
            node.selected = false;
          }

          updateNode(node);
        });
      };



      // if this code is uncommented selection updates will be instantaniously pushed.
      // if (props.graph) {
      //   props.setSelectedNodes(props.graph.nodes.filter((node) => node.selected === true));
      // }


      if (isDragging) {
        checkNodesWithinPolygon();
      }
    };

    const endLassoSelection = (event: FederatedPointerEvent) => {
      viewport.current?.plugins.resume('drag');
      isDragging = false;
      lassoSelection.clear();
      points = []; // Clear the points array

      if (props.graph) {
        props.setSelectedNodes(props.graph.nodes.filter((node) => node.selected === true));
      }
    };

    if (!viewport.current) throw Error('Viewport is undefined');
    lassoSelectionLayer.eventMode = 'dynamic';
    viewport.current.on('mousedown', startLassoSelection);
    viewport.current.on('mousemove', updateLassoSelection);
    viewport.current.on('mouseup', endLassoSelection);
    viewport.current.on('mouseupoutside', endLassoSelection);
  };

  const resetAllSelections = () => {
    if (!props.graph) return;
    props.graph.nodes.forEach((node) => {
      node.selected = false;
    });

    props.setSelectedNodes([]);
    update();
  };

  const setupLayout = () => {
    const layoutFactory = new LayoutFactory();
    layoutAlgorithm = layoutFactory.createLayout(config.LAYOUT_ALGORITHM);

    if (!layoutAlgorithm) throw Error('LayoutAlgorithm is undefined');

    const graphologyGraph = new MultiGraph();
    props.graph?.nodes.forEach((node) => {
      graphologyGraph.addNode(node.id);
    });
    props.graph?.links.forEach((link) => {
      graphologyGraph.addEdge(link.source, link.target);
    });
    layoutAlgorithm.setVerbose(false);

    const boundingBox = { x1: 0, x2: app.renderer.screen.width, y1: 0, y2: app.renderer.screen.height };
    layoutAlgorithm.layout(graphologyGraph, boundingBox);
  };

  return (
    <>
      {popups.map((popup) => (
        <NLPopup onClose={() => {}} data={popup} key={popup.node.id} />
      ))}
      {quickPopup && <NLPopup onClose={() => {}} data={quickPopup} />}
      <div className="h-full w-full overflow-hidden" ref={ref}></div>
    </>
  );
};
